/* 
   psd.c         : Spectrum analysis tool main file
   author        : Ramachandran Ramani

   Copyright (C) 2004-2005 by the University of Southern California
   $Id: psd.c,v 1.13 2005/04/15 01:08:51 fabio Exp $

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License,
   version 2, as published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
*/

#include<stdio.h>
#include<stdlib.h>
#include<unistd.h>
#include<stdarg.h>
#include<math.h>
#include<fftw3.h>

/* Boolean type */
typedef unsigned char bool;
#define false 0;
#define true 1;

struct ele
{
 float val;
 int pos;
};

void usage(char *s)
{
  printf("Usage : ./spect -f data-file [-s sampling-frequency]\n");
  printf("Please view the README file for details\n\n");

  exit(0);
}

int readInputFile(struct ele **x, char *input_filename, char *in_mod,
		  double *sampling_frequency, bool spec_sampl_freq)
{
  FILE *input_samples;  /* Input file for reading samples */
  FILE *f_in; /* Input file when second option is exercised */
  double inc_time; /* Time increments */
  float temp_arr[2]; /* Time Stamp */
  double sampling_time; /* Sampling time */
  int number_of_samples = 0; /* Number of samples in the input file */
  int index = 0, i;
  float tempf, tempf1;

  input_samples = fopen(input_filename, "r"); 

  if(!spec_sampl_freq){

    /* Read the first two samples from the file and determine sampling
       frequency */
    while(fscanf(input_samples, "%f %f", &tempf, &tempf1) != EOF){
      if(index < 2){
	temp_arr[index] = tempf;
	++index; 
      }
      ++number_of_samples;
    }

    /* To find the sampling frequency */
    tempf = temp_arr[1] - temp_arr[0];
    *sampling_frequency = 1.0 / tempf;
  }
  else{
    /* When the sampling frequency is specified, we just need to count
       the samples */
    while(fscanf(input_samples, "%f", &tempf) != EOF)
      ++number_of_samples;

    /* Sampling frequency already specified by user */
    sampling_time = 1.0 / *sampling_frequency;
    inc_time = sampling_time;
  }

  /* Create storage space for reading the input samples */
  *x = (struct ele *) malloc(number_of_samples * sizeof(struct ele));

  fseek(input_samples, 0, 0);

  /* Now we do a second pass and read all samples into our storage */
  if(!spec_sampl_freq){

    for(i = 0; i < number_of_samples; i++){
      (*x)[i].pos = i;
      fscanf(input_samples, "%f %f", &tempf, &(*x)[i].val);
    }
  }
  else{
    f_in = fopen(in_mod, "w");

    for(i = 0; i < number_of_samples; i++)
      {
	(*x)[i].pos = i;
	fscanf(input_samples, "%f", &(*x)[i].val);
	fprintf(f_in, "%f %f\n", inc_time, (*x)[i].val);
	inc_time = inc_time + sampling_time;
      }

    fflush(f_in);
    fclose(f_in);
  }

  fclose(input_samples);

  return(number_of_samples);
}

main(int argc, char *argv[])
{
  /* Variables used for computations */
  int i, j;
  bool flag;
  int number_of_samples = 0;
  double *abval, *cumpsd;
  double sum, freq, quantile_freq;
  float average, std_dev;
  double sampling_frequency = 0.0; /* Sampling frequency */

  /* File pointers */
  FILE *fpsd;  /* Output samples */
  FILE *fcpsd; /* Cumulative spectrum */
  FILE *statistics; /* The statistical details */
  FILE *gn_plot;

  /* Pointers to the various filenames we will open */
  char *input_filename; /* assigning the input trace file name to
			   temporary file */
  char *out_psd;        /* output psd file name */
  char *out_cpsd;       /* output cpsd file name */
  char *out_details;    /* output details file name */
  char *in_mod;         /* output file with data and timestamps, used
			   only when user provides the sampling
			   frequency */

  /* Indicator variables */
  bool spec_sampl_freq = false;

  /* Variables used for command-line parsing */	
  extern char *optarg;
  extern int opterr;
  int option;
  opterr = 0;
	
  /* Input buffer for samples */
  struct ele *x;

  /* FFTW variables */
  fftw_complex *fft_in, *fft_out;
  fftw_plan fft_plan;

  /* Parse command-line options */
  if(argc == 1)
    usage(argv[0]);

  while((option = getopt(argc, argv,"f:s:")) != EOF){

    switch(option) {

    case 'f':

      input_filename = strdup(optarg);

      in_mod = malloc((sizeof(char) * (strlen(input_filename) + 4)));
      sprintf(in_mod, "%s.in", input_filename);

      out_psd = malloc((sizeof(char) * (strlen(input_filename) + 5)));
      sprintf(out_psd, "%s.psd", input_filename);

      out_cpsd = malloc((sizeof(char) * (strlen(input_filename) + 6)));
      sprintf(out_cpsd, "%s.cpsd", input_filename);

      out_details = malloc((sizeof(char) * (strlen(input_filename) + 9)));
      sprintf(out_details, "%s.details", input_filename);

      break;
	
    case 's':

      sampling_frequency = atof(optarg);
      spec_sampl_freq = true;

      break;
	 
    default:

      printf("Invalid option\n");
      usage(argv[0]);
    }
  }

  fpsd = fopen(out_psd, "w");
  fcpsd = fopen(out_cpsd, "w");
  statistics = fopen(out_details, "w");
  gn_plot = popen("gnuplot", "w"); /* Opening a pipe for passing input arguments */

  /* Process input file */
  number_of_samples = readInputFile(&x, input_filename, in_mod,
				    &sampling_frequency, spec_sampl_freq);
     
  /* First calculate the mean value of the number of packets */
  average = 0;

  for(i = 0; i < number_of_samples; i++){
    average = average + x[i].val;
  }

  average = average / (float) number_of_samples;

  /* Then, calculate the standard deviation */
  std_dev = 0;

  for(i = 0; i < number_of_samples; i++)
    std_dev = std_dev + (x[i].val - average) * (x[i].val - average);

  std_dev = std_dev / (float) number_of_samples;
 
  /* Subtracting the mean value of the number of packets */
  for(i = 0; i < number_of_samples; i++)
    x[i].val = x[i].val - average;

  /* Initialize fftw */
  fft_in = fftw_malloc(number_of_samples * sizeof(fftw_complex));
  fft_out = fftw_malloc(number_of_samples * sizeof(fftw_complex));

  for(i = 0; i < number_of_samples; i++){
    fft_in[i][0] = x[i].val;
    fft_in[i][1] = 0;
  }

  /* Calculate */
  fft_plan = fftw_plan_dft_1d(number_of_samples, fft_in, fft_out,
			      FFTW_FORWARD, FFTW_ESTIMATE);
  fftw_execute(fft_plan);

  /* Find out the power spectral density */
  abval = (double *) malloc(number_of_samples / 2 * sizeof(double));

  for(i = 0; i < number_of_samples / 2; i++)
    for(j = 0; j < 1; j++){
      abval[i] = sqrt((fft_out[i][j] * fft_out[i][j]) +
		      (fft_out[i][j+1] * fft_out[i][j+1]));
      freq = i / (float) number_of_samples * (sampling_frequency);
      fprintf(fpsd, "%g %g\n", freq, abval[i]);
    }

  fflush(fpsd);    
	
  /* Now, we find out the cumulative power spectral density */
  cumpsd = (double *) malloc(number_of_samples / 2 * sizeof(double));

  for(i = 0; i < number_of_samples / 2; i++){
    sum = 0;
    for(j = 0; j <= i; j++)
      sum = sum + abval[j];

    cumpsd[i] = sum;
  } 

  /* Write the cumulative psd into cpsd output file and find the 60%
     quantile frequency */
  flag = false;

  for(i = 0; i < number_of_samples / 2; i++){
    cumpsd[i] = cumpsd[i] / (cumpsd[number_of_samples / 2 - 1]);
    freq = i / (float) number_of_samples * (sampling_frequency);
             
    fprintf(fcpsd, "%g %g\n", freq, cumpsd[i]);
	     
    if(!flag){
      if((int)(10 * cumpsd[i]) == 6){
	quantile_freq = freq;
	fprintf(statistics,
		"\nThe 60 percent quantile frequency %g Hz\n",
		quantile_freq);
	flag = true;
      }
    }
  }

  fflush(fcpsd);   

  /* Done, now write statistics */
  fprintf(statistics,
	  "Sampling frequency : %g\n"
	  "Number of Samples : %d\n"
	  "Mean : %g\n"
	  "Standard Deviation : %g\n"
	  "60 percent quantile frequency : %g",
	  sampling_frequency, number_of_samples,
	  average, std_dev, quantile_freq);
	
  if(quantile_freq <= 220)
    fprintf(statistics, "\n\nMultiple source attack", quantile_freq);
  else
    fprintf(statistics, "\n\nSingle source attack", quantile_freq);
 
  fflush(statistics);
  
  if(!spec_sampl_freq){
    fprintf(gn_plot,
	    "set output \"%s.ps\"\n"
	    "set terminal postscript portrait color\n"
	    "set multiplot\n"
	    "set border 3\n"
	    "set xtics nomirror\n"
	    "set ytics nomirror\n\n"
	    "set size 1,0.25\n"
	    "set origin 0,0.75\n"
	    "set xlabel \"Time\"\n"
	    "set ylabel \"Packets/unit time\"\n"
	    "plot \'%s\' notitle w lines\n\n"
	    "set size 1,0.25\n"
	    "set origin 0,0.5\n"
	    "set xlabel \"Frequency\"\n"
	    "set ylabel \"psd\"\n"
	    "plot \'%s\' notitle w lines\n\n"
	    "set size 1,0.25\n"
	    "set origin 0,0.25\n"
	    "set xlabel \"Frequency\"\n"
	    "set ylabel \"cpsd\"\n"
	    "plot \'%s\' notitle w lines",
	    input_filename, input_filename, out_psd, out_cpsd);  
  }
  else{
    fprintf(gn_plot,
	    "set output \"%s.ps\"\n"
	    "set terminal postscript portrait color\n"
	    "set multiplot\n"
	    "set border 3\n"
	    "set xtics nomirror\n"
	    "set ytics nomirror\n\n"
	    "set size 1,0.25\n"
	    "set origin 0,0.75\n"
	    "set xlabel \"Time\"\n"
	    "set ylabel \"Packets/unit time\"\n"
	    "plot \'%s\' notitle w lines\n\n"
	    "set size 1,0.25\n"
	    "set origin 0,0.5\n"
	    "set xlabel \"Frequency\"\n"
	    "set ylabel \"psd\"\n"
	    "plot \'%s\' notitle w lines\n\n"
	    "set size 1,0.25\n"
	    "set origin 0,0.25\n"
	    "set xlabel \"Frequency\"\n"
	    "set ylabel \"cpsd\"\n"
	    "plot \'%s\' notitle w lines",
	    input_filename, in_mod, out_psd, out_cpsd);  
  }

  /* Free memory of buffers */
  free(x);
  free(abval);
  free(cumpsd);

  /* Clean up fftw structures */
  fftw_destroy_plan(fft_plan);
  fftw_free(fft_in);
  fftw_free(fft_out);

  /* Close all files */
  pclose(gn_plot);
  fclose(fpsd);
  fclose(fcpsd);
  fclose(statistics);
}
